package meerkat.bulletinboard.sqlserver;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Crypto.*;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * Created by Arbel Deutsch Peled on 07-Dec-15.
 * This server version allows for reading of messages with several tags and/or signatures
 * The superclass only allows one constraint per type.
 */
public class EnhancedSQLiteBulletinBoardServer extends SQLiteBulletinBoardServer {

    /**
     * This class implements a comparator for the MessageFilter class
     * The comparison is done solely by comparing the type of the filter
     * This is used to sort the filters by type
     */
    public class FilterTypeComparator implements Comparator<MessageFilter>{

        @Override
        public int compare(MessageFilter filter1, MessageFilter filter2) {
            return filter1.getTypeValue() - filter2.getTypeValue();
        }
    }

    @Override
    public BulletinBoardMessageList readMessages(MessageFilterList filterList) throws CommunicationException {

        PreparedStatement pstmt;
        ResultSet messages, signatures;

        long entryNum;
        BulletinBoardMessageList.Builder resultListBuilder = BulletinBoardMessageList.newBuilder();
        BulletinBoardMessage.Builder messageBuilder;

        String sql;
        String sqlSuffix = "";

        List<MessageFilter> filters = new ArrayList<MessageFilter>(filterList.getFilterList());
        int i;

        boolean tagsRequired = false;
        boolean signaturesRequired = false;

        boolean isFirstFilter = true;

        Collections.sort(filters, new FilterTypeComparator());

        // Check if Tag/Signature tables are required for filtering purposes.

        sql = "SELECT MsgTable.EntryNum, MsgTable.Msg FROM MsgTable";

        // Add conditions.

        if (!filters.isEmpty()){
            sql += " WHERE";

            for (MessageFilter filter : filters){

                if (filter.getType().getNumber() != FilterType.MAX_MESSAGES_VALUE){
                    if (isFirstFilter){
                        isFirstFilter = false;
                    } else{
                        sql += " AND";
                    }
                }

                switch (filter.getType().getNumber()){
                    case FilterType.EXACT_ENTRY_VALUE:
                        sql += " MsgTable.EntryNum = ?";
                        break;
                    case FilterType.MAX_ENTRY_VALUE:
                        sql += " MsgTable.EntryNum <= ?";
                        break;
                    case FilterType.MAX_MESSAGES_VALUE:
                        sqlSuffix += " LIMIT = ?";
                        break;
                    case FilterType.MSG_ID_VALUE:
                        sql += " MsgTable.MsgId = ?";
                        break;
                    case FilterType.SIGNER_ID_VALUE:
                        sql += " EXISTS (SELECT 1 FROM SignatureTable"
                                + " WHERE SignatureTable.SignerId = ? AND SignatureTable.EntryNum = MsgTable.EntryNum)";
                        break;
                    case FilterType.TAG_VALUE:
                        sql += " EXISTS (SELECT 1 FROM TagTable"
                                + " INNER JOIN MsgTagTable ON TagTable.TagId = MsgTagTable.TagId"
                                + " WHERE TagTable.Tag = ? AND MsgTagTable.EntryNum = MsgTable.EntryNum)";
                        break;
                }
            }

            sql += sqlSuffix;
        }

        // Make query.

        try {
            pstmt = connection.prepareStatement(sql);

            // Specify values for filters.

            i = 1;
            for (MessageFilter filter : filters){

                switch (filter.getType().getNumber()){

                    case FilterType.EXACT_ENTRY_VALUE: // Go through.
                    case FilterType.MAX_ENTRY_VALUE:
                        pstmt.setLong(i, filter.getEntry());
                        i++;
                        break;

                    case FilterType.MSG_ID_VALUE: // Go through.
                    case FilterType.SIGNER_ID_VALUE:
                        pstmt.setBytes(i, filter.getId().toByteArray());
                        i++;
                        break;

                    case FilterType.TAG_VALUE:
                        pstmt.setString(i, filter.getTag());
                        i++;
                        break;

                    // The max-messages condition is applied as a suffix. Therefore, it is treated differently.
                    case FilterType.MAX_MESSAGES_VALUE:
                        pstmt.setLong(filters.size(), filter.getMaxMessages());
                        i++;
                        break;

                }
            }

            // Run query.

            messages = pstmt.executeQuery();

            // Compile list of messages.

            sql = "SELECT Signature FROM SignatureTable WHERE EntryNum = ?";
            pstmt = connection.prepareStatement(sql);

            while (messages.next()){

                // Get entry number and retrieve signatures.

                entryNum = messages.getLong(1);
                pstmt.setLong(1, entryNum);
                signatures = pstmt.executeQuery();

                // Create message and append signatures.

                messageBuilder = BulletinBoardMessage.newBuilder()
                        .setEntryNum(entryNum)
                        .setMsg(UnsignedBulletinBoardMessage.parseFrom(messages.getBytes(2)));

                while (signatures.next()){
                    messageBuilder.addSig(Signature.parseFrom(signatures.getBytes(1)));
                }

                // Finalize message and add to message list.

                resultListBuilder.addMessage(messageBuilder.build());

            }

            pstmt.close();

        } catch (SQLException e){
            throw new CommunicationException("Error reading messages from DB: " + e.getMessage());
        } catch (InvalidProtocolBufferException e) {
            throw new CommunicationException("Invalid data from DB: " + e.getMessage());
        }

        //Combine results and return.

        return resultListBuilder.build();
    }
}
