package meerkat.bulletinboard;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.List;
import java.util.Random;

import com.google.protobuf.ByteString;

import meerkat.comm.CommunicationException;
import meerkat.crypto.concrete.ECDSASignature;
import meerkat.protobuf.BulletinBoardAPI.BulletinBoardMessage;
import meerkat.protobuf.BulletinBoardAPI.FilterType;
import meerkat.protobuf.BulletinBoardAPI.MessageFilter;
import meerkat.protobuf.BulletinBoardAPI.MessageFilterList;
import meerkat.protobuf.BulletinBoardAPI.UnsignedBulletinBoardMessage;

import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;

public class GenericBulletinBoardServerTest {

	protected BulletinBoardServer bulletinBoardServer;
	private ECDSASignature signers[];
	private ByteString[] signerIDs;
	
	private Random random;

	private static String KEYFILE_EXAMPLE = "/certs/enduser-certs/user1-key-with-password-secret.p12";
	private static String KEYFILE_EXAMPLE3 = "/certs/enduser-certs/user3-key-with-password-shh.p12";
	
	private static String KEYFILE_PASSWORD1 = "secret";
	private static String KEYFILE_PASSWORD3 = "shh";
	
	public static String CERT1_PEM_EXAMPLE = "/certs/enduser-certs/user1.crt";
	public static String CERT3_PEM_EXAMPLE = "/certs/enduser-certs/user3.crt";

	/**
	 * @param bulletinBoardServer is an initialized server.
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws CertificateException
	 * @throws KeyStoreException
	 * @throws NoSuchAlgorithmException
	 * @throws IOException
	 * @throws UnrecoverableKeyException
	 * @throws CommunicationException
	 */
	public void init(BulletinBoardServer bulletinBoardServer) {
		
		this.bulletinBoardServer = bulletinBoardServer;
		
		signers = new ECDSASignature[2];
		signerIDs = new ByteString[signers.length];
        signers[0] = new ECDSASignature();
        signers[1] = new ECDSASignature();
		
		InputStream keyStream = getClass().getResourceAsStream(KEYFILE_EXAMPLE);
        char[] password = KEYFILE_PASSWORD1.toCharArray();

		KeyStore.Builder keyStoreBuilder = null;
		try {
			keyStoreBuilder = signers[0].getPKCS12KeyStoreBuilder(keyStream, password);

		signers[0].loadSigningCertificate(keyStoreBuilder);
        
        signers[0].loadVerificationCertificates(getClass().getResourceAsStream(CERT1_PEM_EXAMPLE));
        
        keyStream = getClass().getResourceAsStream(KEYFILE_EXAMPLE3);
        password = KEYFILE_PASSWORD3.toCharArray();
        
        keyStoreBuilder = signers[1].getPKCS12KeyStoreBuilder(keyStream, password);
        signers[1].loadSigningCertificate(keyStoreBuilder);
        
        signers[1].loadVerificationCertificates(getClass().getResourceAsStream(CERT3_PEM_EXAMPLE));

		} catch (IOException e) {
			System.err.println("Failed reading from signature file " + e.getMessage());
			fail("Failed reading from signature file " + e.getMessage());
		} catch (CertificateException e) {
			System.err.println("Failed reading certificate " + e.getMessage());
			fail("Failed reading certificate " + e.getMessage());
		} catch (KeyStoreException e) {
			System.err.println("Failed reading keystore " + e.getMessage());
			fail("Failed reading keystore " + e.getMessage());
		} catch (NoSuchAlgorithmException e) {
			System.err.println("Couldn't find signing algorithm " + e.getMessage());
			fail("Couldn't find signing algorithm " + e.getMessage());
		} catch (UnrecoverableKeyException e) {
			System.err.println("Couldn't find signing key " + e.getMessage());
			fail("Couldn't find signing key " + e.getMessage());
		}

		// Get signer IDs
		// TODO: remove this after creation of getSignerID method

		UnsignedBulletinBoardMessage msg = UnsignedBulletinBoardMessage.newBuilder().build();

		for (int i = 0 ; i < signers.length ; i++) {

			try {
				signers[i].updateContent(UnsignedBulletinBoardMessage.newBuilder().build());
				signerIDs[i] = signers[i].sign().getSignerId();
			} catch (SignatureException e) {
				System.err.println("Error signing message" + e.getMessage());
				fail("Error signing message" + e.getMessage());
				return;
			}
		}
        
        random = new Random(0); // We use insecure randomness in tests for repeatability
	}
	
	private byte randomByte(){
		return (byte) random.nextInt();
	}
	
	private String randomString(){
		return new BigInteger(130, random).toString(32);
	}

	private final int TAG_NUM = 5;		// Number of tags.
	private final int MESSAGE_NUM = 32;	// Number of messages (2^TAG_NUM).


	private String[] tags;
	private byte[][] data;

	/**
	 * Tests writing of several messages with multiple tags and signatures.
	 * @throws CommunicationException
	 * @throws SignatureException
	 * @throws InvalidKeyException
	 * @throws CertificateException
	 * @throws IOException
     */
	public void testInsert() {

		final int BYTES_PER_MESSAGE_DATA = 50;	// Message size.

		tags = new String[TAG_NUM];
		data = new byte[MESSAGE_NUM][BYTES_PER_MESSAGE_DATA];

		UnsignedBulletinBoardMessage.Builder unsignedMsgBuilder;
		BulletinBoardMessage.Builder msgBuilder;

		int i, j;

		// Generate random data.

		for (i = 1; i <= MESSAGE_NUM; i++) {
			for (j = 0; j < BYTES_PER_MESSAGE_DATA; j++) {
				data[i - 1][j] = randomByte();
			}
		}

		for (i = 0; i < TAG_NUM; i++) {
			tags[i] = randomString();
		}

		// Build messages.

		for (i = 1; i <= MESSAGE_NUM; i++) {
			unsignedMsgBuilder = UnsignedBulletinBoardMessage.newBuilder()
					.setData(ByteString.copyFrom(data[i - 1]));

			// Add tags based on bit-representation of message number.

			int copyI = i;
			for (j = 0; j < TAG_NUM; j++) {
				if (copyI % 2 == 1) {
					unsignedMsgBuilder.addTag(tags[j]);
				}

				copyI >>>= 1;
			}

			// Build message.

			msgBuilder = BulletinBoardMessage.newBuilder()
					.setMsg(unsignedMsgBuilder.build());

			// Add signatures.

			try {

				if (i % 2 == 1) {
					signers[0].updateContent(msgBuilder.getMsg());
					msgBuilder.addSig(signers[0].sign());

					if (i % 4 == 1) {
						signers[1].updateContent(msgBuilder.getMsg());
						msgBuilder.addSig(signers[1].sign());
					}
				}

			} catch (SignatureException e) {
				fail(e.getMessage());
			}

			// Post message.

			try {
				bulletinBoardServer.postMessage(msgBuilder.build());
			} catch (CommunicationException e) {
				fail(e.getMessage());
			}
		}
	}

	/**
	 * Tests retrieval of messages written in {@Link #testInsert()}
	 * Only queries using one tag filter
	 */
	public void testSimpleTagAndSignature(){

		List<BulletinBoardMessage> messages;

		// Check tag mechanism
		
		for (int i = 0 ; i < TAG_NUM ; i++){
			
			// Retrieve messages having tag i

			try {

				messages = bulletinBoardServer.readMessages(
						MessageFilterList.newBuilder()
							.addFilter(MessageFilter.newBuilder()
									.setType(FilterType.TAG)
									.setTag(tags[i])
									.build()
							)
							.build()
					)
					.getMessageList();

			} catch (CommunicationException e) {
				fail(e.getMessage());
				return;
			}

			// Assert that the number of retrieved messages is correct.
			
			assertThat(messages.size(), is(MESSAGE_NUM / 2));
			
			// Assert the identity of the messages.

			for (BulletinBoardMessage msg : messages){
				
				// Assert serial number and raw data.
				
				assertThat((msg.getEntryNum() >>> i) % 2 , is((long) 1));
				assertThat(msg.getMsg().getData().toByteArray(), is(data[(int) msg.getEntryNum() - 1]));
				
				// Assert signatures.

				try {

					if (msg.getEntryNum() % 2 == 1) {
						signers[0].initVerify(msg.getSig(0));
						signers[0].updateContent(msg.getMsg());
						assertTrue("Signature did not verify!", signers[0].verify());

						if (msg.getEntryNum() % 4 == 1) {
							signers[1].initVerify(msg.getSig(1));
							signers[1].updateContent(msg.getMsg());
							assertTrue("Signature did not verify!", signers[1].verify());

							assertThat(msg.getSigCount(), is(2));
						} else {
							assertThat(msg.getSigCount(), is(1));
						}
					} else {
						assertThat(msg.getSigCount(), is(0));
					}
				} catch (Exception e) {
					fail(e.getMessage());
				}
			}
			
		}
		
	}

	/**
	 * Tests retrieval of messages written in {@Link #testInsert()} using multiple tags/signature filters.
	 */
	public void testEnhancedTagsAndSignatures(){

		List<BulletinBoardMessage> messages;
		MessageFilterList.Builder filterListBuilder = MessageFilterList.newBuilder();

		int expectedMsgCount = MESSAGE_NUM;

		// Check multiple tag filters.

		for (int i = 0 ; i < TAG_NUM ; i++) {

			filterListBuilder.addFilter(
					MessageFilter.newBuilder()
						.setType(FilterType.TAG)
						.setTag(tags[i])
						.build()
			);

			try {
				messages = bulletinBoardServer.readMessages(filterListBuilder.build()).getMessageList();
			} catch (CommunicationException e) {
				System.err.println("Failed retrieving multi-tag messages from DB: " + e.getMessage());
				fail("Failed retrieving multi-tag messages from DB: " + e.getMessage());
				return;
			}

			expectedMsgCount /= 2;

			assertThat(messages.size(), is(expectedMsgCount));

			for (BulletinBoardMessage msg : messages) {
				for (int j = 0 ; j < i ; j++) {
					assertThat((msg.getEntryNum() >>> j) % 2, is((long) 1));
				}
			}
		}

		// Check multiple signature filters.

		filterListBuilder = MessageFilterList.newBuilder()
				.addFilter(MessageFilter.newBuilder()
					.setType(FilterType.SIGNER_ID)
					.setId(signerIDs[0])
					.build())
				.addFilter(MessageFilter.newBuilder()
					.setType(FilterType.SIGNER_ID)
					.setId(signerIDs[1])
					.build()
				);

		try {
			messages = bulletinBoardServer.readMessages(filterListBuilder.build()).getMessageList();
		} catch (CommunicationException e) {
			System.err.println("Failed retrieving multi-signature message from DB: " + e.getMessage());
			fail("Failed retrieving multi-signature message from DB: " + e.getMessage());
			return;
		}

		assertThat(messages.size(), is(MESSAGE_NUM / 4));

		for (BulletinBoardMessage message : messages) {
			assertThat(message.getEntryNum() % 4, is((long) 1));
		}

	}
	
	public void close(){
		signers[0].clearSigningKey();
		signers[1].clearSigningKey();
		try {
			bulletinBoardServer.close();
		} catch (CommunicationException e) {
			System.err.println("Error closing server " + e.getMessage());
			fail("Error closing server " + e.getMessage());
		}
	}
}
